local addon = MissionMinder:CreateAddon("MissionMinder-OrderHall")
local module = addon:CreateModule("missionPlanner")

---------------------------------------------
-- LOCALE
---------------------------------------------
local L = MissionMinder.locale
local L_1000 = L["K"]
local L_COMPLETED = L["completed"]
local L_WORKING = L["working"]
local L_IN_PARTY = L["in party"]
local L_AVAILABLE = L["available"]
local L_PARTY = L["Party"]
local L_COUNTERS = L["Counters"]
local L_NO_COUNTERS_AVAILABLE = L["No counters available"]
local L_ASSIGNED_MISSION = L["Assigned mission"]
MISSIONMINDER_MINIMIZE_MISSION = L["Minimize mission"];



---------------------------------------------
-- Mission Planner
---------------------------------------------



--[[local function createInlineTexture(texture)
	return texture and ("|T" .. texture .. ":0|t") or "";
end


local function isOnMission(follower)
	return follower and follower.status == GARRISON_FOLLOWER_ON_MISSION;
end

local function contains(list, value)
	if (list) then
		for _,v in pairs(list) do
			if(v == value) then
				return true
			end
		end
	end
	return false
end]]



--[[

local function ignoreTrait(traitID, hasBonusXPReward)
	if ( options.IgnoreXPGainTraits() and not hasBonusXPReward and expGainTraits[traitID]) then
		return true
	else
		return false
	end
end




local function getMissionRewards(mission)
	local rewards = C_Garrison.GetMissionRewardInfo(mission.missionID);

	local t = {};
	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			local xp = reward.followerXP;
			if ( xp >= 1000 ) then
				xp = xp/1000 .. L_1000;
			end
			tinsert(t, "+" .. xp .. " " .. createInlineTexture(reward.icon));
		elseif (reward.itemID) then
			local itemTexture = select(10, GetItemInfo(reward.itemID));
			if(reward.quantity == 1) then
				tinsert(t, createInlineTexture(itemTexture));
			else
				tinsert(t, reward.quantity .. "x " .. createInlineTexture(itemTexture));
			end
		elseif(reward.currencyID) then
			if (reward.currencyID == 0) then -- gold
				tinsert(t, GetCoinTextureString(reward.quantity));
			else
				tinsert(t, reward.quantity .. " " .. createInlineTexture(reward.icon));
			end
		else
			-- because this is tooltip based, we ignore the value instead of writing the name
		end
	end

	return table.concat(t, ", ");
end

local function MissionHasBonusXPReward(missionID)
	local rewards = C_Garrison.GetMissionRewardInfo(missionID);
	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			return true;
		end
	end

	return false;
end



local function levelComparator(left, right)
	if (left.iLevel ~= right.iLevel ) then
		return left.iLevel > right .iLevel;
	else
		return left.level > right.level;
	end
end

local function getMissionTimeleft(follower)
	if isOnMission(follower) then
		if ( C_Garrison.GetFollowerMissionTimeLeftSeconds(follower.followerID) <= 0 ) then
			return L_COMPLETED, nil, nil, nil; -- Default yellow
		else
			local timeleft = C_Garrison.GetFollowerMissionTimeLeft(follower.followerID);
			return string.format("(%s)", timeleft), 0.7, 0.7, 0.7; -- Gray
		end
	elseif (follower.status == GARRISON_FOLLOWER_WORKING) then
		return L_WORKING, 1, 0, 0; -- red
	elseif (follower.status == GARRISON_FOLLOWER_IN_PARTY) then
		return L_IN_PARTY -- use default color
	else
		return L_AVAILABLE, 0, 1, 0; -- Green
	end
end



local cache = {
	["isDirty"] = true,
	["abilities"] = {},
	["traits"] = {}
};


local function dirtyMissionCounterCache()
	cache.isDirty = true;
end

local function getFollowerCountersForMisison(followerID, missionID)
	if( cache.isDirty or not cache.abilities[missionID] or not cache.traits[missionID] ) then
		cache.abilities[missionID] = C_Garrison.GetBuffedFollowersForMission(missionID);
		cache.traits[missionID] = C_Garrison.GetFollowersTraitsForMission(missionID);
	end

	local abilities = cache.abilities[missionID];
	local traits = cache.traits[missionID];

	local hasBonusXP = MissionHasBonusXPReward(missionID);

	local counters = {};
	local hasNonFilterable = false
	if ( abilities and abilities[followerID] ) then
		for i = 1, #abilities[followerID] do
			local ability = abilities[followerID][i];
			counters[#counters + 1] = createInlineTexture(ability.icon);

			hasNonFilterable = true
		end
	end

	if ( traits and traits[followerID] ) then
		for i = 1, #traits[followerID] do
			local trait = traits[followerID][i];
			counters[#counters + 1] = createInlineTexture(trait.icon);

			if ( not ignoreTrait(trait.traitID, hasBonusXP) ) then
				hasNonFilterable = true
			end
		end
	end

	if ( hasNonFilterable ) then
		return counters;
	else
		return nil;
	end
end

local function createMissionTooltip(self, button)
	local mission = self.info;

	if ( not mission or not self:IsShown() ) then -- Not a valid button for mouseover
		return
	end

	if(mission.inProgress) then -- Do not show tooltip on inProgress missions
		return;
	end

	--[===[@debug@
	GameTooltip:AddLine("MissionID = " .. mission.missionID);
	--@end-debug@]===]
	GameTooltip:AddLine(" ");

	local party = nil;
	if ( minimizedParties[mission.missionID] ) then
		party = minimizedParties[mission.missionID]
		local hasParty = false;

		for i=1,#party do
			if ( party[i] ) then
				if (not hasParty ) then
					GameTooltip:AddLine(L_PARTY);
					hasParty = true;
				end

				local follower = C_Garrison.GetFollowerInfo(party[i]);
				local r1, g1, b1 = getFollowerMissionBiasRGB(mission.missionID, follower.followerID);
				local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);

				local level = follower.level
				if (mission.iLevel and mission.iLevel > 600 and follower.level == GARRISON_FOLLOWER_MAX_LEVEL) then
					level = follower.iLevel
				end

				local left = string.format("[%d] %s %s", level, follower.name, counters and table.concat(counters, "") or "");
				GameTooltip:AddLine(left, r1, g1, b1);
			end
		end

		if (hasParty) then
			GameTooltip:AddLine(" ");
		end
	end


	local followers = C_Garrison.GetFollowers();
	local numFollowers = followers and #followers or 0;
	table.sort(followers, levelComparator);

	local counterAvailable = false;
	for i1 = 1, numFollowers do -- Necessary to walk all followers because we want to sort by level
		local follower = followers[i1];

		if ( MissionMinder_FollowerCanBeSentOnMission(follower) and not contains(party, follower.followerID)) then
			local r1, g1, b1 = getFollowerMissionBiasRGB(mission.missionID, follower.followerID);

			local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);
			if ( counters and #counters > 0 ) then
				if ( not counterAvailable ) then
					GameTooltip:AddLine(L_COUNTERS);
					counterAvailable = true;
				end

				local level = follower.level
				if (mission.iLevel and mission.iLevel > 600 and follower.level == GARRISON_FOLLOWER_MAX_LEVEL) then
					level = follower.iLevel
				end

				local left = string.format("[%d] %s %s", level, follower.name, table.concat(counters, ""));
				local right, r2, g2, b2 = getMissionTimeleft(follower);
				GameTooltip:AddDoubleLine(left, right, r1, g1, b1, r2, g2, b2);
			end
		end
	end

	if ( not counterAvailable ) then
		GameTooltip:AddLine(L_NO_COUNTERS_AVAILABLE);
	end

	GameTooltip:Show(); -- resize the tooltip for the new lines
end

local function createFollowersTooltip(self)
	local follower = C_Garrison.GetFollowerInfo(self.id);
	if ( not follower.isCollected ) then
		return;
	end


	GameTooltip:SetOwner(self, "ANCHOR_NONE");
	GameTooltip:SetPoint("TOPRIGHT", self, "TOPLEFT", -10);
	local r, g, b = GetItemQualityColor(follower.quality);
	GameTooltip:SetText(follower.name, r, g, b);

	--[===[@debug@
	GameTooltip:AddLine("FollowerID = " .. follower.followerID);
	if ( GarrisonMissionFrame.MissionTab.MissionPage:IsShown() ) then
		local mission = GarrisonMissionFrame.MissionTab.MissionPage.missionInfo;
		local bias = C_Garrison.GetFollowerBiasForMission(mission.missionID, follower.followerID)
		GameTooltip:AddLine(string.format("Bias = %d%%", bias * 100));
	end
	--@end-debug@]===]

	if ( isOnMission(follower) ) then
		GameTooltip:AddLine(getMissionTimeleft(follower))
	end
	GameTooltip:AddLine(" ");


	local availableMissions = GarrisonMissionFrame.MissionTab.MissionList.availableMissions;
	C_Garrison.GetAvailableMissions(availableMissions);
	table.sort(availableMissions, levelComparator);

	if ( followerMissions[follower.followerID] ) then
		local mission =  followerMissions[follower.followerID]

		local r, g, b = getFollowerMissionBiasRGB(mission.missionID, follower.followerID);
		local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);
		local level = mission.level
		if (mission.iLevel and mission.iLevel > 600) then
			level = mission.iLevel
		end

		local rewards = getMissionRewards(mission);
		local left = string.format("[%d] %s (%s)", level, mission.name, rewards);
		local right = counters and table.concat(counters, "") or nil;

		GameTooltip:AddLine(L_ASSIGNED_MISSION);
		GameTooltip:AddDoubleLine(left, right, r, g, b, 1, 1, 1);
		GameTooltip:AddLine(" ");
	end

	local atleastOneMission = false;
	for _,mission in pairs(availableMissions) do
		local r, g, b = getFollowerMissionBiasRGB(mission.missionID, follower.followerID);

		local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);
		if ( counters and #counters > 0 ) then
			if( not atleastOneMission ) then
				GameTooltip:AddLine(GARRISON_MISSIONS);
				atleastOneMission = true;
			end

			local level = mission.level
			if (mission.iLevel and mission.iLevel > 600) then
				level = mission.iLevel
			end

			local rewards = getMissionRewards(mission);
			local left = string.format("[%d] %s (%s)", level, mission.name, rewards);
			local right = table.concat(counters, "");
			GameTooltip:AddDoubleLine(left, right, r, g, b, 1, 1, 1);
		end
	end

	if ( not atleastOneMission ) then
		GameTooltip:AddLine(L_NO_COUNTERS_AVAILABLE);
	end

	GameTooltip:Show();
	cache.isDirty = false;
end







local function hideMissionCountersTooltip(self)
	GameTooltip:Hide();
end]]



--[[local followerDragSource = nil;
local function setDragSource(self, button)
	if (GarrisonFollowerPlacer:IsShown()) then
		DEBUG("SetDragSource(%s)", tostring(self));
		followerDragSource = self;
	end
end

local function clearDragSource()
DEBUG("ClearDragSource()");
	followerDragSource = nil;
end

local function getFollowerFrameIndex(self)
	local partyFollowers = GarrisonMissionFrame.MissionTab.MissionPage.Followers;
	for i = 1, #partyFollowers do
		if ( partyFollowers[i] == self) then
			return i
		end
	end

	return 0;
end

local function setFollower(destFrame, destInfo, ...)
	DEBUG("SetFollower(%s, %s) DragSource=%s", tostring(destFrame), destInfo.followerID, tostring(followerDragSource));
	if ( followerDragSource ) then
		local srcFrame = followerDragSource;
		local srcInfo = destFrame.info
		srcIndex = getFollowerFrameIndex(srcFrame);
		destIndex = getFollowerFrameIndex(destFrame);

		DEBUG("srcIndex=%d, destIndex=%d", srcIndex, destIndex)

		if (srcIndex > destIndex ) then
			-- We always want to assign the "lower" frame first to keep the follower order consistent
			--There are some internals to garrison missions that can cause weirdness for the user.
			--like hidden enemy modifiers on [615] Auchindoom!
			srcFrame, destFrame = destFrame, srcFrame
			srcInfo, destInfo = destInfo, srcInfo
		elseif (srcIndex < destIndex ) then
			-- Must first unassign dest follower
			GarrisonMissionPage_ClearFollower(destFrame);
		else -- Targetting the same frame, only do one set (destination)
			srcInfo = nil;
		end

		if(srcInfo) then
			DEBUG("src SetFollower(%s, %s)", tostring(srcFrame), srcInfo.followerID);
			module.GarrisonMissionPage_SetFollower(srcFrame, srcInfo, ...)
		end
		if (destInfo) then
			DEBUG("dest SetFollower(%s, %s)", tostring(destFrame), destInfo.followerID);
			module.GarrisonMissionPage_SetFollower(destFrame, destInfo, ...)
		end
	else
		module.GarrisonMissionPage_SetFollower(destFrame, destInfo, ...)
	end
end


]]--


--[[function setFollowerFromFrame(frame, followerID)
	local missionPage = GarrisonMissionFrame.MissionTab.MissionPage
	if ( not followerID ) then
		GarrisonMissionPage_ClearFollower(frame);
		return;
	elseif (frame.info) then
		GarrisonMissionPage_ClearFollower(frame);
	end

	local info = C_Garrison.GetFollowerInfo(followerID);
	frame.info = info;
	frame.Name:Show();
	frame.Name:SetText(info.name);
	if (frame.Class) then
		frame.Class:Show();
		frame.Class:SetAtlas(info.classAtlas);
	end
	frame.PortraitFrame.Empty:Hide();

	local showItemLevel;
	if ( missionPage.showItemLevel and info.level == GarrisonMissionFrame.followerMaxLevel ) then
		frame.PortraitFrame.LevelBorder:SetAtlas("GarrMission_PortraitRing_iLvlBorder");
		frame.PortraitFrame.LevelBorder:SetWidth(70);
		showItemLevel = true;
	else
		frame.PortraitFrame.LevelBorder:SetAtlas("GarrMission_PortraitRing_LevelBorder");
		frame.PortraitFrame.LevelBorder:SetWidth(58);
		showItemLevel = false;
	end
	GarrisonMissionFrame_SetFollowerPortrait(frame.PortraitFrame, info, showItemLevel);

	local counters = GarrisonMissionFrame.followerCounters and GarrisonMissionFrame.followerCounters[frame.info.followerID] or nil;
	if (counters) then
		for i = 1, #counters do
			if (not frame.Counters[i]) then
				frame.Counters[i] = CreateFrame("Frame", nil, frame, "GarrisonMissionAbilityLargeCounterTemplate");
				frame.Counters[i]:SetPoint("LEFT", frame.Counters[i-1], "RIGHT", 16, 0);
			end
			local Counter = frame.Counters[i];
			Counter.info = counters[i];
			Counter.info.showCounters = true;
			Counter.Icon:SetTexture(counters[i].icon);
			Counter.tooltip = counters[i].name;
			Counter:Show();
		end
		for i = (#counters + 1), #frame.Counters do
			frame.Counters[i]:Hide();
		end
	end

	-- update follower list
	GarrisonMissionFrame.followerTraits = C_Garrison.GetFollowersTraitsForMission(missionPage.missionInfo.missionID);
	GarrisonFollowerList_UpdateFollowers(GarrisonMissionFrame.FollowerList);

	frame.PortraitFrame.Level:SetTextColor(getFollowerMissionBiasRGB(missionPage.missionInfo.missionID, info.followerID));
	GarrisonMissionPage_UpdateMissionForParty();

	if ( missionPage.missionInfo.numFollowers == 1 ) then
		local model = missionPage.FollowerModel;
		model:Show();
		model:SetTargetDistance(0);
		GarrisonMission_SetFollowerModel(model, info.followerID, info.displayID);
		model:SetHeightFactor(info.height or 1);
		model:InitializeCamera(info.scale or 1);
		model:SetFacing(-.2);
		missionPage.EmptyFollowerModel:Hide();
	end

	GarrisonMissionPage_SetCounters();
end

function MissionMinderOrderHall_MinimizeMission(self, ...)
	local frame = self:GetParent():GetParent()
	local missionPage = frame:GetMissionPage()
	local mission = missionPage.missionInfo
	if( not mission ) then
		return
	end

	if ( not minimizedParties[mission.missionID] ) then
		minimizedParties[mission.missionID] = {}
	end
	local party = minimizedParties[mission.missionID]

	for i = 1, #missionPage.Followers do
		local followerFrame = missionPage.Followers[i];
		local followerID = clearFollowerFromFrame(followerFrame);
		party[i] = followerID;
		if ( followerID ) then
			followerMissions[followerID] = mission
		end

		module:DEBUG(string.format("Saving follower [%d][%d] = %s", mission.missionID, i, tostring(party[i])));
	end

	--MissionMinder_MissionFrame_UpdatePartyLevelUp(missionFrame);
	--updatePartyXPGains();
	--missionPage.FollowerModel:Hide();
	--GarrisonMissionPage_UpdateEmptyString();

	OrderHallMissionFrame:GetMissionPage():Hide();
	OrderHallMissionFrame.MissionTab.MissionList:Show();
	OrderHallMissionFrame.followerCounters = nil;
	OrderHallMissionFrame.followerTraits = nil;
	OrderHallMissionFrame:GetMissionPage().missionInfo = nil;
end

local function determineCountersFromMinimizedParty(missionID)
	local threats = {};
	threats.full = {};
	threats.partial = {};
	threats.away = {};
	threats.worker = {};


	if ( not minimizedParties[missionID] ) then
		minimizedParties[missionID] = {}
	end
	local party = minimizedParties[missionID]

	for i = 1, #party do
		local followerID = party[i];
		if ( followerID ) then
			local follower = C_Garrison.GetFollowerInfo(followerID);

			if ( follower.isCollected and follower.status ~= GARRISON_FOLLOWER_INACTIVE ) then
				local bias = C_Garrison.GetFollowerBiasForMission(missionID, followerID);
				if ( bias > -1.0 ) then
					local abilities = C_Garrison.GetFollowerAbilities(followerID);
					for j = 1, #abilities do
						for counterMechanicID in pairs(abilities[j].counters) do
							local isFullCounter = C_Garrison.IsMechanicFullyCountered(missionID, followerID, counterMechanicID, abilities[j].id);
							local map = ( isFullCounter ) and threats.full or threats.partial;
							map[counterMechanicID] = (map[counterMechanicID] or 0) + 1
						end
					end
				end
			end
		end
	end

	return threats;
end

local nextFollowerListUpdate = nil;
local function followerList_OnUpdate(self)
	local now = GetTime();
	if ( not nextFollowerListUpdate or now > nextFollowerListUpdate ) then
		nextMissionListUpdate = math.floor(now + 0.2) -- Update every 0.2 seconds
		GarrisonFollowerList_Update(GarrisonMissionFrame)
	end
end

local statusPriority = {
	[GARRISON_FOLLOWER_IN_PARTY] = 1,
	[GARRISON_FOLLOWER_WORKING] = 2,
	[GARRISON_FOLLOWER_ON_MISSION] = 3,
	[GARRISON_FOLLOWER_EXHAUSTED] = 4,
	[GARRISON_FOLLOWER_INACTIVE] = 5,
}

local function sortFollowersForMission(self)
	local followers = self.followers;
	local followerCounters = GarrisonMissionFrame.followerCounters;
	local followerTraits = GarrisonMissionFrame.followerTraits;
	local checkAbilities = followerCounters and followerTraits and GarrisonMissionFrame.MissionTab:IsVisible();
	local mentorLevel, mentorItemLevel = 0, 0;
	if ( checkAbilities ) then
		mentorLevel = GarrisonMissionFrame.MissionTab.MissionPage.mentorLevel or 0;
		mentorItemLevel = GarrisonMissionFrame.MissionTab.MissionPage.mentorItemLevel or 0;
	end

	-- Load followers in the current mission
	local MissionPage = GarrisonMissionFrame.MissionTab.MissionPage
	local hasMission = GarrisonMissionFrame.MissionTab:IsVisible() and MissionPage.missionInfo
	local party = {}
	if ( hasMission ) then
		local missionparty = { MissionPage.Follower1, MissionPage.Follower2, MissionPage.Follower3 }
		for i = 1, #missionparty do
			if ( missionparty[i].info ) then
				party[missionparty[i].info.followerID] = true
			end
		end
	end

	local comparison = function(index1, index2)
		local follower1 = followers[index1];
		local follower2 = followers[index2];

		if ( follower1.isCollected ~= follower2.isCollected ) then
			return follower1.isCollected;
		end

		-- treat IN_PARTY status as no status
		local status1 = follower1.status;
		if ( status1 == GARRISON_FOLLOWER_IN_PARTY and (not hasMission or party[follower1.followerID]) ) then
			status1 = nil;
		end
		local status2 = follower2.status;
		if ( status2 == GARRISON_FOLLOWER_IN_PARTY and (not hasMission or party[follower2.followerID]) ) then
			status2 = nil;
		end
		if ( status1 and not status2 ) then
			return false;
		elseif ( not status1 and status2 ) then
			return true;
		end

		if ( status1 ~= status2 ) then
			return statusPriority[status1] < statusPriority[status2];
		end
		-- sorting: level > item level > (num counters for mission) > (num traits for mission) > quality > name
		local follower1Level = max(follower1.level, mentorLevel);
		local follower2Level = max(follower2.level, mentorLevel);
		if ( follower1Level ~= follower2Level ) then
			return follower1Level > follower2Level;
		end
		local follower1ItemLevel = max(follower1.iLevel, mentorItemLevel);
		local follower2ItemLevel = max(follower2.iLevel, mentorItemLevel);
		if ( follower1Level == GARRISON_FOLLOWER_MAX_LEVEL and follower1ItemLevel ~= follower2ItemLevel ) then		-- only checking follower 1 because follower 2 has same level at this point
			return follower1ItemLevel > follower2ItemLevel;
		end
		if ( checkAbilities and not status1 and follower1.isCollected ) then		-- only checking follower 1 because follower 2 has same status and collected-ness at this point
			local numCounters1 = followerCounters[follower1.followerID] and #followerCounters[follower1.followerID] or 0;
			local numCounters2 = followerCounters[follower2.followerID] and #followerCounters[follower2.followerID] or 0;
			if ( numCounters1 ~= numCounters2 ) then
				return numCounters1 > numCounters2;
			end
			local numTraits1 = followerTraits[follower1.followerID] and #followerTraits[follower1.followerID] or 0;
			local numTraits2 = followerTraits[follower2.followerID] and #followerTraits[follower2.followerID] or 0;
			if ( numTraits1 ~= numTraits2 ) then
				return numTraits1 > numTraits2;
			end
		end

		if ( follower1.quality ~= follower2.quality ) then
			return follower1.quality > follower2.quality;
		end
		return strcmputf8i(follower1.name, follower2.name) < 0;
	end

	table.sort(self.followersList, comparison);
end

local IN_PARTY_BUSY_COLOR = { 0, 0.2, 0, 0.44 }

local function updateInPartyFollowerBusyFrameColor()
	local buttons = GarrisonMissionFrame.FollowerList.listScroll.buttons;

	for i = 1, #buttons do
		local button = buttons[i];
		if ( button:IsShown() and button.info and button.info.status == GARRISON_FOLLOWER_IN_PARTY ) then
			button.BusyFrame.Texture:SetTexture(unpack(IN_PARTY_BUSY_COLOR))
		end
	end
end





local function updateMissionButtonsForParty()
	local self = GarrisonMissionFrame.MissionTab.MissionList;

	local buttons = self.listScroll.buttons
	for i = 1, #buttons do
		local button = buttons[i]
		local mission = button.info

		if ( mission ) then
			local totalTimeString, totalTimeSeconds, isMissionTimeImproved, successChance, partyBuffs, isEnvMechanicCountered, xpBonus, materialMultiplier, goldMultiplier = C_Garrison.GetPartyMissionInfo(mission.missionID)

			-- Mission duration
			if ( not self.showInProgress and isMissionTimeImproved ) then
				button.Summary:SetFormattedText(PARENS_TEMPLATE, GREEN_FONT_COLOR_CODE..totalTimeString..FONT_COLOR_CODE_CLOSE)
			end

			-- Rewards
			GarrisonMissionPage_UpdateRewardQuantities(button, materialMultiplier, goldMultiplier);
		end
	end
end]]--

---------------------------------------------
-- INITIALIZE
---------------------------------------------
function module.initialize()
	local missionFrame = OrderHallMissionFrame
	local missionPage = missionFrame:GetMissionPage()

--[[module:RegisterEvent("GARRISON_FOLLOWER_LIST_UPDATE", dirtyMissionCounterCache);
	module:RegisterEvent("GARRISON_FOLLOWER_XP_CHANGED", dirtyMissionCounterCache);
	module:RegisterEvent("GARRISON_FOLLOWER_REMOVED", dirtyMissionCounterCache);

	-- Mission tooltips
	local missionList = GarrisonMissionFrame.MissionTab.MissionList
	local missionButtons = missionList.listScroll.buttons;
	for i = 1, #missionButtons do
		local button = missionButtons[i]
		button:HookScript("OnEnter", createMissionTooltip);
	end
	hooksecurefunc("GarrisonMissionButton_OnEnter", createMissionTooltip)

	-- Periodically update mission tooltips
	for i = 1, #missionButtons do
		missionButtons[i].UpdateTooltip = GarrisonMissionButton_OnEnter
	end

	-- Update mission tooltip mechanics with minimized party counters
	GarrisonMission_DetermineCounterableThreats = determineCountersFromMinimizedParty
]]
	-- Load minimized mission party
	-- MissionList is initialized early, no need to hook OrderHallMissionFrame:SetupMissionList()
	local missionButtons = missionFrame.MissionTab.MissionList.listScroll.buttons or {}
	for i = 1, #missionButtons do
		local button = missionButtons[i]
		button:HookScript("OnClick", MissionMinder_LoadSavedParty);
	end

	-- Show mission expiration
	for i = 1, #missionButtons do
		local button = missionButtons[i]
		button.Expires = button:CreateFontString("MissionExpires", "ARTWORK", "GameTooltipText")
		button.Expires:SetPoint("BOTTOM", button.MissionType);
		button.Expires:Show();
	end
	hooksecurefunc(missionFrame.MissionTab.MissionList, "Update", MissionMinder_AvailableMissionList_ShowTimeRemaining)

	-- Show minimized parties on mission button
	hooksecurefunc(missionFrame.MissionTab.MissionList, "Update", MissionMinder_AvailableMissionList_ShowMinimizedParty);

	-- Minimize mission when right click on MissionPage
	missionPage:SetScript("OnClick", function() missionPage.MinimizeButton:Click(); end);

	-- Minimize mission on FollowerFrame RightClick
	local partyFollowers = missionPage.Followers;
	for i = 1, #partyFollowers do
		partyFollowers[i]:SetScript("OnMouseUp", MissionMinder_MissionPageFollowerFrame_OnMouseUp)
	end

	-- Follower tooltips
	--[[local followerButtons = FollowerList.listScroll.buttons;
	for i = 1, #followerButtons do
		local button = followerButtons[i]
		button:SetScript("OnEnter", createFollowersTooltip);
		button:SetScript("OnLeave", hideMissionCountersTooltip);
	end]]

	-- Double click to assign followers to mission
	missionPage:HookScript("OnHide", MissionMinder_FollowerList_DisableSingleClickTimer)
	MissionMinder_FollowerList_Initialize(missionFrame.FollowerList, function()
		local followerButtons = missionFrame.FollowerList.listScroll.buttons
		for i = 1, #followerButtons do
			local button = followerButtons[i].Follower
			button.OriginalOnClick = button:GetScript("OnClick")
			button:SetScript("OnClick", MissionMinder_FollowerList_SingleClick);
			button:SetScript("OnDoubleClick", MissionMinder_FollowerList_AssignFollowerToMission);
		end
	end)

	-- Show remaining XP to level on follower list button
	hooksecurefunc(missionFrame.FollowerList, "UpdateData", MissionMinder_FollowerList_UpdateXPRemaining)
	MissionMinder_FollowerList_Initialize(missionFrame.FollowerList, function()
		local followerButtons = missionFrame.FollowerList.listScroll.buttons
		for i = 1, #followerButtons do
			local button = followerButtons[i].Follower
			local xpRemaining = button:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall");
			xpRemaining:SetPoint("BOTTOMRIGHT", button.BG);
			xpRemaining:SetText(nil);
			xpRemaining:Show();
			button.xpRemaining = xpRemaining;
		end
	end)

	-- Color follower levels by bias
	hooksecurefunc(missionFrame.FollowerList, "UpdateData", MissionMinder_FollowerList_UpdateBiasColorFollowerLevels)

	-- Change In Party follower busy frame color
	--[[hooksecurefunc("GarrisonFollowerList_Update", updateInPartyFollowerBusyFrameColor)

	-- Drag-to-swap party members
	module.GarrisonMissionPage_SetFollower = GarrisonMissionPage_SetFollower
	GarrisonMissionPage_SetFollower = setFollower
	GarrisonFollowerPlacerFrame:HookScript("OnClick", clearDragSource);
	for i = 1, #partyFollowers do
		partyFollowers[i]:HookScript("OnDragStart", setDragSource);
		partyFollowers[i]:HookScript("OnReceiveDrag", clearDragSource);
	end
	for i = 1, #followerButtons do
		local button = followerButtons[i]
		button:HookScript("OnDragStart", clearDragSource)
	end]]

	-- Follower level up
	hooksecurefunc(missionFrame.FollowerList, "UpdateData", MissionMinder_FollowerList_UpdateFollowerLevelUp)
	hooksecurefunc(missionFrame, "AssignFollowerToMission", MissionMinder_MissionFrame_UpdatePartyLevelUp);
	hooksecurefunc(missionFrame, "RemoveFollowerFromMission", MissionMinder_MissionFrame_UpdatePartyLevelUp);
	MissionMinder_FollowerList_Initialize(missionFrame.FollowerList, function()
		local followerButtons = missionFrame.FollowerList.listScroll.buttons
		for i = 1, #followerButtons do
			local button = followerButtons[i].Follower
			local levelUp = button.PortraitFrame:CreateTexture();
			levelUp:SetPoint("BOTTOM", button.PortraitFrame, "BOTTOMLEFT")
			levelUp:Hide();
			button.levelUp = levelUp
		end
	end)
	for i = 1, #partyFollowers do
		local levelUp = partyFollowers[i].PortraitFrame:CreateTexture();
		levelUp:SetPoint("BOTTOM", partyFollowers[i].PortraitFrame, "BOTTOMLEFT")
		levelUp:Hide();
		partyFollowers[i].levelUp = levelUp
	end

	-- Estimate party member XP gains
	hooksecurefunc(missionFrame, "AssignFollowerToMission", MissionMinder_MissionPage_UpdatePartyXPGains);
	hooksecurefunc(missionFrame, "RemoveFollowerFromMission", MissionMinder_MissionPage_UpdatePartyXPGains);
	for i = 1, #partyFollowers do
		local xpGain = partyFollowers[i]:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
		xpGain:SetPoint("TOP", partyFollowers[i], "BOTTOM")
		xpGain:SetTextColor(0, 1, 0)
		xpGain:SetText(nil)
		xpGain:Show()
		partyFollowers[i].XPGain = xpGain
	end

	-- Periodically update follower list (for mission time remaining)
	--[[FollowerList:SetScript("OnUpdate", followerList_OnUpdate);]]

	-- Show reward quality borders
--	hooksecurefunc("GarrisonMissionButton_SetRewards", MissionButton_updateRewardQualityBorder);
	hooksecurefunc("GarrisonLandingPageReportList_UpdateItems", MissionMinder_LandingPageReport_AvailableMissions_UpdateRewardQualityBorder);
	GarrisonLandingPageReport.List.listScroll.update = GarrisonLandingPageReportList_UpdateItems
	--hooksecurefunc("GarrisonMissionPage_SetReward", updateRewardQualityBorder);

	-- Clear all parties when mission frame closes
	missionFrame:HookScript("OnHide", MissionMinder_MissionFrame_ClearAllParties)

	-- Change order of followers to better aid mission planning
--[[	module.GarrisonFollowerList_SortFollowers = GarrisonFollowerList_SortFollowers
	GarrisonFollowerList_SortFollowers = sortFollowersForMission

	-- Update mission buttons for party
	hooksecurefunc("GarrisonMissionList_Update", updateMissionButtonsForParty)

	-- Update mission button rewards to include currencyQuantity
	hooksecurefunc("GarrisonMissionButton_SetRewards", setRewardCurrencyQuantity)
]]

	missionPage:SetScript("OnKeyUp", MissionMinder_MissionPage_SpaceBarStartMission)

	--[[local debug = function(...) end
	module:RegisterEvent("GARRISON_LANDINGPAGE_SHIPMENTS", debug);
	module:RegisterEvent("GARRISON_SHOW_LANDING_PAGE", debug);
	module:RegisterEvent("GARRISON_HIDE_LANDING_PAGE", debug);
	module:RegisterEvent("GARRISON_BUILDING_ACTIVATABLE", debug);
	module:RegisterEvent("GARRISON_BUILDING_ACTIVATED", debug);
	module:RegisterEvent("GARRISON_ARCHITECT_OPENED", debug);
	module:RegisterEvent("GARRISON_MISSION_FINISHED", debug);
	module:RegisterEvent("GARRISON_MISSION_NPC_OPENED", debug);
	module:RegisterEvent("GARRISON_SHIPYARD_NPC_OPENED", debug);
	module:RegisterEvent("GARRISON_INVASION_AVAILABLE", debug);
	module:RegisterEvent("GARRISON_INVASION_UNAVAILABLE", debug);
	module:RegisterEvent("SHIPMENT_UPDATE", debug);
	module:RegisterEvent("ZONE_CHANGED_NEW_AREA", debug);
	module:RegisterEvent("ZONE_CHANGED", debug);
	module:RegisterEvent("ZONE_CHANGED_INDOORS", debug);
	module:RegisterEvent("GARRISON_FOLLOWER_CATEGORIES_UPDATED", debug)]]
end
