local addon = MissionMinder:CreateAddon("MissionMinder")
local module = addon:CreateModule("skipFightAnimations")


---------------------------------------------
-- LOCALE
---------------------------------------------
local L = MissionMinder.locale


---------------------------------------------
-- Data
---------------------------------------------
local SKIP_ANIMATION_CONTROL = {
	[1] = MissionMinder.SkipFightAnimations,	-- line between encounters
	[2] = MissionMinder.SkipFightAnimations,	-- check that models are loaded
	[3] = MissionMinder.SkipFightAnimations,	-- model fight
	[4] = MissionMinder.SkipFightAnimations,	-- impact sound when follower hits
	[5] = MissionMinder.SkipFightAnimations,	-- X over portrait
	[6] = MissionMinder.SkipFightAnimations,	-- evaluate whether to do next encounter or move on
	[7] = nil,									-- reward panel
	[8] = nil,									-- explode the lock if mission successful
	[9] = nil,									-- show all the mission followers
	[10] = nil									-- follower xp
};


-- Skip any fight animations and sound as defind in SKIP_ANIMATION_CONTROL
local function skipAnimations(self, elapsed, ...)
	if ( MissionMinder.SkipFightAnimations() ) then
		local skipControl = SKIP_ANIMATION_CONTROL[self.animIndex];
		while ( skipControl and skipControl() ) do
			self.animTimeLeft = 0;

			skipControl = SKIP_ANIMATION_CONTROL[self.animIndex + 1];
			if(skipControl and skipControl()) then
				-- only increase animation index if the next one is ALSO skipped. This is necessary because the original
				--OnUpdate function will increase animation index for us becaue animTimeLeft <= 0
				--The result is that leaving this loop, self.animIndex == 6, but will be increased to 7 for us.
				self.animIndex = self.animIndex + 1;
			end
		end
	end

	self.Original_OnUpdate(self, elapsed, ...);
end


-- if SkipFightAnimations, hide enemy unit frames from completion frame
local function hideEncounterFrame(self)
	if ( MissionMinder.SkipFightAnimations() ) then
		self.MissionComplete.Stage.EncountersFrame:Hide();
	end
end


-- Show mission success/failure message
local function showSuccessFailureMessage(self, missionID, canComplete, succeeded, overmaxSucceeded, followerDeaths)
	if ( MissionMinder.SkipFightAnimations() and self.currentMission and self.currentMission.missionID == missionID ) then
		self.BonusRewards.Saturated:Show();
		self.BonusRewards.Saturated.FadeIn:Play();
		if ( --[[self.currentMission.]]succeeded ) then
			self.ChanceFrame.ResultText:SetText(GARRISON_MISSION_SUCCESS);
			self.ChanceFrame.ResultText:SetTextColor(0.1, 1, 0.1);
			self.ChanceFrame.ResultAnim:Play();
			self.BonusRewards.ChestModel:SetAnimation(0, 0);
			PlaySound("UI_Garrison_CommandTable_MissionSuccess_Stinger");
		else
			self.ChanceFrame.ResultText:SetText(GARRISON_MISSION_FAILED);
			self.ChanceFrame.ResultText:SetTextColor(1, 0.1, 0.1);
			self.ChanceFrame.ResultAnim:Play();
			self.NextMissionButton:Enable();
			PlaySound("UI_Garrison_Mission_Complete_MissionFail_Stinger");
		end
	end
end


function module.register(missionFrame)
	local missionComplete = missionFrame.MissionComplete

	-- Skip fight animations
	-- INSECURE HOOK - Requires pre-hook to override the local animation control
	missionComplete.Original_OnUpdate = missionComplete.OnUpdate
	missionComplete.OnUpdate = skipAnimations;
	if(missionFrame.MissionComplete:GetScript("OnUpdate")) then
		missionFrame.MissionComplete:SetScript("OnUpdate", missionComplete.OnUpdate);
	end

	-- Hide encounter frame (enemy unit frames)
	hooksecurefunc(missionFrame, "MissionCompleteInitialize", hideEncounterFrame);

	-- Show the success/failure message
	hooksecurefunc(missionComplete, "OnMissionCompleteResponse", showSuccessFailureMessage);
end
